from SimulatorFramework import TileAssembly
import RobinsonTiles
from SimulatorFramework import SimUtils
import os
import ast
import random


class RobinsonAssembly(TileAssembly.Assembly):
    def __init__(self):
        TileAssembly.Assembly.__init__(self)

        self.growthPattern = []
        self.growthPatternLoc = 0
        self.growthTileLocations = [(0,0)]

    def ClearSpace(self):
        TileAssembly.Assembly.ClearSpace(self)
        self.growthPatternLoc = 0
        self.growthTileLocations = [(0,0)]
        
    def AddToFrontierIfNecessary(self, coords):
        fittingNbrSet = self.GetFittingTilesForLoc(coords)
        if fittingNbrSet is None:
            return
        elif len(fittingNbrSet) == 1:
            fLoc = RobinsonTiles.RobinsonFrontierLocation(coords, fittingNbrSet)
            self.frontier.discard(fLoc)
            self.frontier.add(fLoc)
        #elif len(fittingNbrSet) > 1 and fittingNbrSet.issubset(self.growthTileTypes):
        elif (len(fittingNbrSet) > 1) and (self.IsSetOfGrowthTiles(fittingNbrSet) == True):
            SimUtils.PrintDebug("Adding a growth tile frontier location at " + str(coords), SimUtils.MSG_MSG)
            fLoc = RobinsonTiles.RobinsonFrontierLocation(coords, fittingNbrSet)
            self.frontier.add(fLoc)
        #print("fitting nbr set: " + str(fittingNbrSet))
        
    def IsSetOfGrowthTiles(self, tileTypes):
        for tt in tileTypes:
            if tt.startswith("growth") == False:
                return False
        return True
    
    def GetFittingTilesForLoc(self, coords):
        retSet = None
        count = 0
        for n in TileAssembly.NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrTile = self.GetLoc(nbrCoords, True)
            if not nbrTile is None:
                oppDir = (count + 4) % 8
                if retSet is None:
                    retSet = nbrTile.type.matches[oppDir]
                else:
                    retSet = retSet.intersection(nbrTile.type.matches[oppDir])
            count = count + 1
        return retSet
    
    def StepForward(self):
        # Pick a spot from the frontier, add a tile
        if len(self.frontier) == 0:
            SimUtils.globalApp.OutputMessage("The frontier is empty!")
            return None
        else:
            frontierLoc = random.sample(self.frontier, 1)[0]
            self.frontier.remove(frontierLoc)
            if frontierLoc.isGrowthLoc == True:
                if self.growthPatternLoc >= len(self.growthPattern):
                    typeName = frontierLoc.GetRandomFittingTileTypeName()
                else:
                    pattern = self.growthPattern[self.growthPatternLoc]
                    matchingTypeNames = []
                    for ttName in frontierLoc.fittingTypeNames:
                        if pattern in ttName:
                            matchingTypeNames.append(ttName)
                    if len(matchingTypeNames) == 0:
                        SimUtils.globalApp.OutputMessage("ERROR: No fitting growth tile at " + str(frontierLoc.coords) + " matches pattern '" + str(pattern) + "'!  Available set was " + str(frontierLoc.fittingTypeNames))
                        self.growthPatternLoc = self.growthPatternLoc + 1
                        return None
                    else:
                        SimUtils.globalApp.OutputMessage("The growth tiles for location " + str(frontierLoc.coords) + " which match '" + str(pattern) + "' are these " + str(len(matchingTypeNames)) + ": " + str(matchingTypeNames))
                        typeName = random.sample(matchingTypeNames, 1)[0]
                        self.growthPatternLoc = self.growthPatternLoc + 1
                if self.AddTile(self.tileTypes[typeName], frontierLoc.coords) == True:
                    self.growthTileLocations.append(frontierLoc.coords)
                    SimUtils.globalApp.OutputMessage("Added a growth tile of type " + typeName + " to location " + str(frontierLoc.coords) + " from " + str(len(frontierLoc.fittingTypeNames)) + " choices " + str(frontierLoc.fittingTypeNames))
                    return (frontierLoc.coords, typeName)
            else:
                typeName = frontierLoc.GetRandomFittingTileTypeName()
                if self.AddTile(self.tileTypes[typeName], frontierLoc.coords) == True:
                    SimUtils.PrintDebug("Added a tile of type " + typeName + " to location " + str(frontierLoc.coords), SimUtils.MSG_MSG)
                    return (frontierLoc.coords, typeName)
            SimUtils.globalApp.OutputMessage("Error adding tile to location" + str(frontierLoc.coords))
            return None
        
    def LoadTileSystem(self, fileName):
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        #try:
        os.stat(fileName)
        rFile = open(fileName, "r")
        SimUtils.globalApp.OutputMessage("Using tile system definition file '" + fileName + "'")
        lines = rFile.readlines()
        rFile.close()
        
        self.systemDefFile = fileName
        self.tileSystemDef = ast.literal_eval(lines[0])
        self.tileSystemDef['MainFile'] = fileName
        self.tileSystemDef['DirName'] = os.path.dirname(fileName)
        
        print("Read tile system " + str(self.tileSystemDef))

        self.tileTypes = dict()
        self.seedPatch = []
        #self.growthTileTypes = []
        self.frontier = set()
        
        fileName = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['NeighborList']
        SimUtils.globalApp.OutputMessage("Reading neighbor list file '" + fileName + "'")
        os.stat(fileName)
        rFile = open(fileName, "r")
        lines = rFile.readlines()
        rFile.close()
        
        #for i in range(len(lines)):
        i = 0
        while i < len(lines):
            name = lines[i][:-3]
            i = i + 2
            
            dirs = []
            for j in range(8):
                #dir = lines[i].strip()
                i = i + 1

                matches = lines[i].strip()
                i = i + 2
                matches = matches.split()
                matchSet = set(matches)
                dirs.append(matchSet)

            i = i + 2
            #tileType = {"name":name, "matches":dirs}
            name = name.replace('"', '')
            img = "TileImage_" + name + ".tif"
            fullPath = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['TilePictureDir'] + "/" + img
            if os.path.isfile(fullPath) == True:
                tileType = RobinsonTiles.RobinsonTile(name, dirs, fullPath)
                self.tileTypes[name] = tileType
            else:
                SimUtils.globalApp.OutputMessage("WARNING:  Missing image file: '" + img + "'")
                tileType = RobinsonTiles.RobinsonTile(name, dirs, None)
                self.tileTypes[name] = tileType
            #print("tileType: " + name)
        SimUtils.globalApp.OutputMessage("Loaded tile set with size: " + str(len(self.tileTypes)))

        fileName = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['SeedPatch']
        SimUtils.globalApp.OutputMessage("Reading seed patch file '" + fileName + "'")
        os.stat(fileName)
        rFile = open(fileName, "r")
        lines = rFile.readlines()
        rFile.close()
        seedDefn = ""
        for line in lines:
            seedDefn = seedDefn + line.strip()
        self.LoadSeed(seedDefn)
        
        return True
    
    def LoadSeed(self, seedDefn):
        seedDefn = seedDefn.replace("{", "[")
        seedDefn = seedDefn.replace("}", "]")
        seedArray = []
        if len(seedDefn) > 0:
            bException = True
            while bException == True:
                try:
                    seedArray = eval(seedDefn)
                    bException = False
                except NameError as e:
                    errStr = str(e)
                    errArr = errStr.split("\'")
                    name = errArr[1]
                    seedDefn = seedDefn.replace(name, "\'" + name + "\'")
                    bException = True

        self.seedPatch = []
        for entry in seedArray:
            name = entry[0]
            x = int(entry[1][0][0])
            y = int(entry[1][0][1])
            orient = entry[1][1]
            name = name + "_" + str(orient)
            self.seedPatch.append((name, (x,y)))
        self.ClearSpace()
        self.frontier = set()
        self.PlaceSeed()

    def LoadSeedOrig(self, seedDefn):
        chars = ' {}\'\"'
        for line in seedDefn:
            entry = line.split(',')
            name = entry[0].strip(chars)
            x = int(entry[1].strip(chars))
            y = int(entry[2].strip(chars))
            orient = entry[3].strip(chars)
            name = name + "_" + orient
            self.seedPatch.append((name, (x,y)))
        self.PlaceSeed()
